#!/usr/bin/env python
"""An actor to process APO status messages and send them to the hub"""

from twisted.internet import reactor

import actorcore.Actor
import actorcore.CmdrConnection

from apoActor import __version__


class APO(actorcore.Actor.Actor):
    def __init__(self, name, productName=None, configFile=None, debugLevel=30):
        actorcore.Actor.Actor.__init__(
            self,
            name,
            productName=productName,
            configFile=configFile,
        )

        self.version = __version__

        self.logger.setLevel(debugLevel)
        self.logger.propagate = True

        #
        # Get filesystems for diskspace check in apoCmd.doStatus
        #
        self.fshost = self.config["diskspace"]["fshost"].split(",")
        self.fsvolumes = self.config["diskspace"]["fsvolumes"].split(",")

        #
        # Get the keys for use in apoCmd.doStatus
        #
        import opscore.protocols.keys as protoKeys

        self.kdict = protoKeys.KeysDictionary.load(name)
        #
        # Get the dictionary for use in talking to the TCC
        #
        self.tccNameDict = {
            "truss25m": "sectrusstemp",
            "airTempPT": "airtemp",
            "winds": "wspeed",
            "windd": "wdir",
            "pressure": "pressure",
            "humidPT": "humid",
        }

        # Initialize the data cache, for use in apoCmd.doStatus
        #
        self.cachedValues = {}

        # Schedule an update.
        #
        self.statusTimer = reactor.callLater(
            int(self.config[self.name]["updateInterval"]),
            self.periodicStatus,
        )
        # Finally start the reactor
        #
        self.run()

    def periodicStatus(self):
        """Run 'apoCmd.doStatus' periodically"""
        #
        # Obtain and send the data
        #
        self.callCommand("update")
        self.statusTimer = reactor.callLater(
            int(self.config[self.name]["updateInterval"]),
            self.periodicStatus,
        )


# To work
if __name__ == "__main__":
    apo = APO("apo", "apoActor", debugLevel=15)



#!/usr/bin/env python

import os
from time import sleep

from opscore.utility.qstr import qstr

from apoActor.wx import WX


class apoCmd(object):
    """Wrap commands to the apo actor"""

    def __init__(self, actor):
        self.actor = actor

        # Declare commands

        self.vocab = [
            ("ping", "", self.ping),
            ("status", "", self.status),
            ("update", "", self.update),
        ]

    def ping(self, cmd):
        """Query the actor for liveness/happiness."""

        cmd.finish("text='Present and (probably) correct'")

    def status(self, cmd):
        """Report status and version; obtain and send current data"""

        self.actor.sendVersionKey(cmd)
        self.doStatus(cmd, flushCache=True)

    def update(self, cmd):
        """Report status and version; obtain and send current data"""

        self.doStatus(cmd=cmd)

    def doStatus(self, cmd=None, flushCache=False):
        """Obtain data and send it to the hub and optionally the TCC"""

        if not cmd:
            cmd = self.actor.bcast

        #
        # Flush the data cache if requested (all data, not just updated data)
        #
        if flushCache:
            self.actor.cachedValues = {}

        w = WX().recvcat("tuples", dict=True)

        #
        # Check for OK reply
        #
        while w.v.recv_status != "OK":
            w = WX().recvcat("tuples", dict=True)
            if w.v.recv_status != "OK":
                cmd.warn('recv_status="%s"' % w.v.recv_status)
                sleep(1)

        #
        # Retrieve file system information
        #
        fsdict = {}
        for volume in self.actor.fsvolumes:
            fsinfo = os.statvfs(volume)
            fssize = fsinfo.f_frsize * fsinfo.f_blocks
            fsfree = fsinfo.f_frsize * fsinfo.f_bavail
            fsdict[volume] = (fssize, fsfree)

        # cmd.debug('text="fsvolumes: %s"' % fsdict)

        #
        # Send updated values to tron
        #
        updated = 0
        new = 0
        keyStrings = []
        for key in self.actor.kdict.keys.values():
            name = key.name
            #
            # Check if key is present in data dictionary
            #
            if name in w.v.d:
                msg = "%s=%s" % (name, w.v.d[name][1])
                #
                # Check for previous presence of key
                #
                if name in self.actor.cachedValues:
                    #
                    # Check for change of timestamp and value
                    #
                    if w.v.d[name] != self.actor.cachedValues[name]:
                        keyStrings.append(msg)
                        updated += 1
                else:
                    keyStrings.append(msg)
                    new += 1

        #
        # Seperately generate disk space keywords for each volume
        #
        for host in self.actor.fshost:
            oneGig = 1024 * 1024 * 1024.0
            for volume in self.actor.fsvolumes:
                ratio = fsdict[volume][1] / fsdict[volume][0]
                msg = "diskspace=%s,%s,%d,%d" % (
                    qstr(host),
                    qstr(volume),
                    fsdict[volume][0] / oneGig,
                    fsdict[volume][1] / oneGig,
                )
                if ratio <= 0.1:
                    cmd.warn(msg)
                else:
                    cmd.inform(msg)

        keyMsg = "; ".join(keyStrings)
        self.actor.logger.info(
            "%d %d %s" % (new, updated, keyMsg if keyMsg else "no updated wx data")
        )

        #
        # Send weather data to TCC if requested
        #
        # The TCC uses pressure in Pascals and humidity as a fraction of one
        # so we convert from inHg -> Pascals and % -> fraction respectively.
        #
        # Note that the TCC only accepts strings up to 132 characters.
        #
        if int(self.actor.config["apo"]["sendToTCC"]):
            new = 0
            tccStrings = []
            for name in self.actor.tccNameDict.keys():
                #
                # Check if key is present in data dictionary
                #
                if name in w.v.d:
                    new += 1
                    #
                    # Get the TCC's name for this key
                    #
                    tccName = self.actor.tccNameDict[name]
                    #
                    # Handle unit conversions
                    #
                    if name == "pressure":
                        tccStrings.append(
                            "%s=%.3f"
                            % (tccName, 3386.3788 * float(w.v.d["pressure"][1]))
                        )
                    elif name == "humidPT":
                        tccStrings.append(
                            "%s=%.3f" % (tccName, 0.01 * float(w.v.d["humidPT"][1]))
                        )
                    else:
                        tccStrings.append("%s=%s" % (tccName, w.v.d[name][1]))

            tccMsg = "set weather " + ", ".join(tccStrings)
            self.actor.logger.info("%d %s" % (new, tccMsg))

            cmdVar = self.actor.cmdr.call(actor="tcc", cmdStr=tccMsg, timeLim=3.0)
            if cmdVar.didFail:
                cmd.warn("text='Failed to set TCC weather'")

        # Store data to use in check for updated data on the next call
        #
        self.actor.cachedValues = w.v.d
        #
        # Finish and send data message to hub
        #
        cmd.finish(keyMsg)



#!/usr/bin/env python

import ctypes
from binascii import crc32
from socket import AF_INET, SOCK_DGRAM, socket
from sys import stdout
from time import sleep
from traceback import print_exc


class WX:
    """Weather server communications"""

    class v:
        """contains values received from the weather server"""

        pass

    def __init__(self, server="10.25.1.254", port=6251, timeout=15):
        self.server = (server, port)
        self.size = 2048
        self.defaultTimeout = timeout

    def connect(self):
        global sock
        try:
            sock = socket(AF_INET, SOCK_DGRAM)
            sock.settimeout(2)
            sock.connect(self.server)
            sock.settimeout(self.defaultTimeout)
        except:
            print_exc()
            return -1
        return 0

    def close(self):
        try:
            sock.close()
        except:
            print_exc()
            return -1
        return 0

    def recv(self):
        """receive a message from the server"""
        try:
            return sock.recv(self.size).decode()
        except:
            print_exc()
            return ""

    def recvonce(self):
        """receive a single message and close the connection"""
        self.connect()
        reply = self.recv().decode()
        self.close()
        return reply

    def send(self, messages):
        """send a list of messages to the server and return any reply"""
        for message in messages:
            try:
                sock.send(message.encode())
            except:
                print_exc()
                return ""

        try:
            return sock.recv(self.size).decode()
        except:
            print_exc()
            return ""

    def sendonce(self, messages):
        """send a list of messages and close the connection"""
        self.connect()
        reply = self.send(messages)
        self.close()
        return reply

    def uint(self, x):
        """converts a signed integer to an unsigned integer"""
        return ctypes.c_uint(x).value

    def recvcat(self, category="all", dict=False, debug=False):
        """recv the category, checking the datagram and
        returning the variables and their values

        format: <d_len> <d_crc32> <d_nval> timeStamp=<n> <keyword>=<val> end
        """

        reply = self.sendonce([category])

        if debug:
            print("category=%s, dict=%d, debug=%d" % (category, dict, debug))
            print(reply)
            stdout.flush()

        if "end" != reply[-3:]:
            self.v.recv_status = "Premature end of data"
            return WX()
        else:
            start = reply.find("timeStamp")
            stop = reply.find(" end")
            tmp = reply[:start].split()  # [len, crc32, nval]

            reply_bytes = reply[start:].encode()

            if -1 == start:
                self.v.recv_status = '"timeStamp" key not found'
            elif -1 == stop:
                self.v.recv_status = '"end" key not found'
            elif int(tmp[0]) != len(reply[start:]):
                self.v.recv_status = "Short packet"
            elif int(tmp[1], 16) != self.uint(0xFFFFFFFF & crc32(reply_bytes)):
                self.v.recv_status = "Invalid CRC32"
            elif int(tmp[2]) != len(reply[start:stop].split()):
                self.v.recv_status = "Missing value(s)"
            else:
                if dict:
                    self.v.d = {}
                    if "tuples" == category:
                        exec(
                            "self.v.d['" +
                            reply[start:stop]
                            .replace("=", "']=")
                            .replace(" ", "; self.v.d['")
                        )
                    else:
                        exec(
                            "self.v.d['" +
                            reply[start:stop]
                            .replace("=", "']='")
                            .replace(" ", "'; self.v.d['") +
                            "'"
                        )

                else:
                    exec(
                        "self.v." +
                        reply[start:stop]
                        .replace("=", "='")
                        .replace(" ", "'; self.v.") +
                        "';"
                    )
                self.v.recv_status = "OK"

        if debug:
            print(start, stop, tmp, int(tmp[0]), "%X" % int(tmp[1], 16), int(tmp[2]))
            print(
                len(reply[start:]),
                "%X" % self.uint(0xFFFFFFFF & crc32(reply[start:])),
                len(reply[start:stop].split()),
            )
            stdout.flush()

        return WX()

    def vclear(self, dict=False):
        """Delete the variables for class v"""
        if dict:
            del self.v
        else:
            for key in vars(self.v).copy().keys():
                exec("del self.v.%s" % key)


if __name__ == "__main__":

    if 0:
        w = WX()
        w.v.recv_status = "OK"
        while w.v.recv_status == "OK":
            w.recvcat(category="all", debug=1)
            sleep(1)
    else:
        print("Variables:\n")
        w = WX().recvcat(debug=1)
        print("recv_status =", w.v.recv_status)
        print(len(vars(w.v)))
        print(vars(w.v))
        w.vclear()
        print(len(vars(w.v)))

        w = WX().recvcat("site", debug=1)
        print("recv_status =", w.v.recv_status)
        print(len(vars(w.v)))

        print("\nDictionary:\n")
        w = WX().recvcat(dict=1, debug=1)
        print("recv_status =", w.v.recv_status)
        print(w.v.d)
        WX().vclear()
        w = WX().recvcat("site", dict=True, debug=True)
        print("recv_status =", w.v.recv_status)
        print(w.v.d)

        print("\nDictionary with Tuples:\n")
        w = WX().recvcat("tuples", dict=1, debug=1)
        print("recv_status =", w.v.recv_status)
        print(w.v.d)

# Traceback (most recent call last):
# File "./wx.py", line 171, in ?
# 	w = WX().recvcat ('tuples', dict=1, debug=1)
# File "./wx.py", line 128, in recvcat
# 	print start, stop, tmp, int (tmp[0]), '%X' % int (tmp[1], 16), int (tmp[2])
# UnboundLocalError: local variable 'start' referenced before assignment
